import sys
import os
import time
import datetime
from cdo import *
from Scientific.N import *
from Scientific.IO.NetCDF import *
from dateutil.relativedelta import *
import numpy
from common import *
class Data:
    """This classe define the set of data for the observation or model
    on which will be calculate the statistics.
    The variables are the following:
    - datanames: names of experiments or observational set (string)
    - varname: name of the variables on which the statitics must be calculated (string)
    - inputdir: directory where are stored the original data (string)
    - outpudir: directory to write the output data (string)
    - filelist: list of file on which the calculations must be done (list of string)
    - sdates: list of startdate associated to each file (list of date)
    - filetype: type of the files in Filelist (string)
    - freq: frequency between 2 timesteps (timedelta)
    - nmenber: number of member (int)"""
    def __init__(self, dataname, varname):
        self._dataname=dataname
        self._varname=varname
        self._inputdir=""
        self._outputdir=""
        self._filelist=[]
        self._sdates=[]
        self._filetype=""
        self._freq=-1
        self._nmember=-1    
        
    
    ######## set ########

    def set_dataname(self, dataname):
        self._dataname=dataname

    def set_varname(self, varname):
        self._varname=varname
   
    def set_inputdir(self, inputdir):
        self._inputdir=inputdir

    def set_outputdir(self, outputdir):
        self._outputdir=outputdir

    def set_filelist(self, filelist):
        self._filelist=[]
        self._filelist.extend(filelist)

    def set_sdates(self, sdates):
        self._sdates=[]
        self._sdates.extend(sdates)

    def set_filetype(self, filetype):
        self._filetype=filetype

    def set_freq(self,freq):
        self._freq=freq

    def set_nmember(self, nmember):
        self._nmember = nmember

    ######## get ########

    def get_dataname(self):
        return self._dataname

    def get_varname(self):
        return self._varname
   
    def get_inputdir(self):
        return self._inputdir

    def get_outputdir(self):
        return self._outputdir

    def get_filelist(self):
        return self._filelist

    def get_sdates(self):
        return self._sdates

    def get_filetype(self):
        return self._filetype

    def get_freq(self):
        return self._freq

    def get_nmember(self):
        return self._nmember          

    def ready(self, date1, date2):
        """ Make the Data object ready to calculate statistics """
        #self.set_inputdirAuto() #look for the output directory
        self.set_outputdirAuto() #look for the output directory
        #self.set_files(date1, date2)
        self.format_files_nc() #set file calendar and frequency
        self.see()

    ######## print ########

    def see(self):
        """ print the data object"""
        print "dataname: ", self.get_dataname()
        print "varname: ", self.get_varname()
        print "inputdir: ", self.get_inputdir()
        print "outputdir: ", self.get_outputdir()
        print "filelist: ", self.get_filelist()
        print "sdates: ", self.get_sdates()
        print "filetype: ", self.get_filetype()
        print "freq: ", self.get_freq()
        print "nmember: ", self.get_nmember()


    def duplicate(self):
        """duplicate all atribute execpt the lists"""
        exp2=Data("","")
        exp2.set_varname(self.get_varname())
        exp2.set_dataname(self.get_dataname())
        exp2.set_freq(self.get_freq())
        exp2.set_inputdir(self.get_inputdir())
        exp2.set_outputdir(self.get_outputdir())
        exp2.set_nmember(self.get_nmember())
        return exp2

    def copy(self, aux):
        """copy all atributes of an object in self"""
        self.set_varname(aux.get_varname())
        self.set_dataname(aux.get_dataname())
        self.set_freq(aux.get_freq())
        self.set_inputdir(aux.get_inputdir())
        self.set_outputdir(aux.get_outputdir())
        self.set_nmember(aux.get_nmember())
        self.set_filelist(list(aux.get_filelist()))
        self.set_sdates(list(aux.get_sdates()))


    ######## set_outputdirAuto ########

    def set_outputdirAuto(self):
        """set the output directory based from 
        the input directory"""
        inputdir=self.get_inputdir()
        print inputdir
        if "/" in  inputdir:
            #remove the last part of the input directory and replace it by statistics
            outputdir="/".join(inputdir.split("/")[:-3])+"/monthly_statistics/"
        else:
            print "The path specify must be an abolute path"
            print "this path is not correct or is empty:"
            print inputdir
            sys.exit(0)
        self.set_outputdir(outputdir)
        if (not os.path.isdir(outputdir)):
            os.mkdir(outputdir)
            print "Created directory:"
            print outputdir


   ######## select_leadtime ########

    def select_leadtime(self, leadtime1, leadtime2):
        """select the time period between leadtime1 and leadtime2
        in all files of the Data object:
        -leadtime1: in month (relativedelta)
        -leadtime2: in month (relativedelta)
        the file must have the correct format for cdo
        (run the format_files_nc method in case of failure)
        (to be changed when the right named convention has been set)
        """
        cdo=Cdo()
        #get the needed variables from the object
        filelist=self.get_filelist()
        sdates=self.get_sdates()
        dataname=self.get_dataname()
        varname=self.get_varname()
        freq=self.get_freq()
        #timedelta can not be compared: find a solution
        #if ((leadtime1!=relativedelta()) and (leadtime2 < leadtime1)):
        #        print leadtime1
        #        print leadtime2
        #        print "the second leadtime must be larger than the first one"
        #        sys.exit(0)

        filelistout=[]
        sdatesout=[]
        
        lead1str=str(leadtime1.months).zfill(2)
        lead2str=str(leadtime2.months).zfill(2)
        for i in range(len(filelist)):
            f=filelist[i]
            d=sdates[i]
            
            fdate1=d+leadtime1
            fdate2=d+leadtime2
            

            print "the selected time period is:"
            print fdate1, fdate2

            #calculate the last date of the file
            lastdate=get_last_date(f)

            #check if the time period is smaller the the length of the file
            
            if (fdate2 > lastdate):
                print fdate2
                print lastdate
                print "the selected time period is not include in the file:"
                print f
                sys.exit(0)


            fileout=(f.split("/")[-1])
            fileout=fileout.split(".")[0]+"_"+lead1str+"-"+lead2str+".nc"
            cdo.seldate(fdate1,fdate2, input=f, output=fileout, options="-r -f nc", Force=False)
            filelistout.append(fileout)
            sdatesout.append(fdate1)
            
        
        
        exp_season=self.duplicate()
        exp_season.set_varname(varname + "_" + lead1str + "-" + lead2str)
        exp_season.set_filelist(list(filelistout[:]))
        exp_season.set_sdates(sdatesout[:])
        return exp_season


    def format_files_nc(self):
        """this fonction check if the file in the filelist 
        have a proper format.
        get the variable freq
        the calendar is checked and set if needed
        the variables incompatibles with cdo are removed"""
        self.see()
        cdo=Cdo()

        #get the needed variables from the object
        filelist=self.get_filelist()
        sdates=self.get_sdates()
        dataname=self.get_dataname()
        varname=self.get_varname()
        
        
        #caculate the frequency and check in all files 
        for i in range(len(filelist)):
            f=filelist[i]
            d=sdates[i]

            #calculate  freq
            
            freq = get_freq(f)

            #test if the taxis axis is well defined
            # add more test if needed
            if freq == 0:
                freq=self.get_freq()
                f=settaxis(dataname, varname, f, d, freq)
                
            if get_last_date(f) <= d:
                f=settaxis(dataname, varname, f, d, freq)
            #if freq was equal too -1 set the new frequency
            if  (self.get_freq() == -1):
                self.set_freq(freq)
            elif (self.get_freq() != freq) and (self.get_freq() != -1):
                print "the file %s does not have a %s frequency" % (f, str(freq))
                sys.exit(0)

            #add here solution for cdo imcompatibilities

            filelist[i]=f

        self.set_freq(freq)
        self.set_filetype("nc")
        self.see()

    def create_monthly_list(self):
        """create a list of Data object
        containing only one month"""
        #get the needed variables from the object
        filelist=list(self.get_filelist())
        sdates=list(self.get_sdates())
        dataname=self.get_dataname()
        freq=self.get_freq()
        
        f=filelist[0]
        lastdate=get_last_date(f)
        nbmon=diff_month(lastdate,sdates[0])
        listmondata=[]
        for imon in range(nbmon+1):
            leadtime1=relativedelta(months=+imon)
            leadtime2=relativedelta(months=+imon+1, days=-1)
            listmondata.append(self.select_leadtime(leadtime1, leadtime2))
        return listmondata




    ###################################
    ########    STATISTICS     ########
    ###################################     


    ######## monthly_percentile ########
 
    def monthly_percentile(self, percentile):
        """calculate the monthly percentile for all file in filelist"""

        cdo=Cdo()
        
        #get the needed variables from the object
        filelist=list(self.get_filelist())
        sdates=list(self.get_sdates())
        dataname=self.get_dataname()
        varname=self.get_varname()
        freq=self.get_freq()
        
        for i in range(len(filelist)):
            f=filelist[i]
            fileout=(f.split("/")[-1])
            fileout=fileout.split(".")[0]+"_q"+str(percentile)+".nc"
            cdo.monpctl(percentile, input=f+" -monmin "+f+" -monmax "+f, 
                        output=fileout,
                        options="-r -f nc")

            filelist[i] = fileout
            #when ncdf4 instaled change name of variables in files


        exp_stat=self.duplicate()
        exp_stat.set_filelist(filelist)
        exp_stat.set_sdates(sdates)
        exp_stat.set_freq = relativedelta(months=+1)
        exp_stat.set_varname(varname + "_q" + str(percentile))
        
        return exp_stat

     ######## monthly_nbday ########

    def monthly_nbday(self, percentile, comp):
        """calculate the monthly percentage of days 
        greater the climatological percentile 
        for all file in filelist.
        Note: the climatological percentile 
        is calculated with all files of filelist
        """
        listmondata=self.create_monthly_list()
        listdatastat=[]
        for data in listmondata:
            datastat=data.seasonal_nbday(percentile, comp)
            listdatastat.append(datastat)

        var0=(listdatastat[0]).get_varname()
        var0=quit_leadtime(var0)
        var0=var0.split("-")
        var1=(listdatastat[-1]).get_varname()
        var1=quit_leadtime(var1)
        var1=var1.split("-")

        varnew=var0[0]+"-"+var1[1]
        expstat=concat_monthly_data(listdatastat,
                                    varnew)
        return expstat

     ######## seasonal_nbday ########


    def seasonal_nbday(self, percentile, comp):
        """calculate the monthly percentile for all file in filelist"""
        cdo=Cdo()
        
        #get the needed variables from the object
        filelist=list(self.get_filelist())
        sdates=list(self.get_sdates())
        dataname=self.get_dataname()
        varname=self.get_varname()
        freq=self.get_freq()
        nmember=self.get_nmember()

        
        #calculate the climatological percentile
        
        #check if there is no time overlap between the file
        for i in range(len(sdates)-1):
            lastdate = get_last_date(filelist[i])
            if (lastdate > sdates[i+1]):
                sdates[i]=sdates
                print "the file are overlaping"
                sys.exit(0)
        merge=cdo.mergetime(input=filelist, options="-O -r -f nc")
        
        
        #open merge file
        ncdfmer=NetCDFFile(merge, "r")
        variables=ncdfmer.variables.keys()
        allDimNames = ncdfmer.dimensions.keys() 

            
        if varname in variables:
            varmer = ncdfmer.variables[varname]
            varaux=varname
        else:
            varaux=varname.split("_")[0]
            if varaux in variables:
                varmer = ncdfmer.variables[varaux]
            else:
                print "the variables %s or %s is not found in the file:" %varname,varaux
                print merge
                sys.exit(0)

        shapmer=shape(varmer)
        if not(len(shapmer) == 4):
            if (len(shapmer) == 3):
                varmer=numpy.resize(varmer,(shapmer[0],1,shapmer[1],shapmer[2]))
                shapmer=shape(varmer)
            else:
                print "the merged file does not have 3 or 4 dimensions"
                sys.exit(0)
        if not(shapmer[1] ==  nmember):
            print "the number of member in the merged file is not correct"
            sys.exit(0)

        print "the merge of file has been done successfully"
        varmer=numpy.resize(varmer,(shapmer[0]*shapmer[1],shapmer[2],shapmer[3]))
        percent=numpy.percentile(varmer, percentile, axis=0)
        filelistout=[]
        for f in filelist:
            print "treating file:", f
            ncdfin=NetCDFFile(f, "r")
            varin=ncdfin.variables[varaux]
            shapin=shape(varin)
            leadtime=shapin[0]
            if not(len(shapin) == 4):
                if not(len(shapin) == 4):
                    if (len(shapin) == 3):
                        varin=numpy.resize(varin,(shapin[0],1,shapin[1],shapin[2]))
                        shapin=shape(varin)
                    else:
                        print "the file does not have 3 or 4 dimensions"
                        sys.exit(0)
            if not(shapin[1] ==  nmember):
                print "the number of member in the file is not correct"
                sys.exit(0)

            varmercomp=numpy.array([percent,]*nmember)
            varout=numpy.zeros((shapmer[1],shapmer[2],shapmer[3]), dtype='f')

            for t in range(leadtime):
                if comp == "greater":
                    varout=varout+numpy.greater_equal(varin[t,:,:,:],varmercomp)
                if comp == "less":
                    varout=varout+numpy.less_equal(varin[t,:,:,:],varmercomp)

            varout=varout/leadtime
            if numpy.amax(varout) > 1:
                print "the percent of day is greater than 1"
                print shape(varout)
                print leadtime
                sys.exit(0)
            #convert the output in a standart python type
            fileout=f.split("/")[-1]
            fileout=fileout.split(".")[0]+"_nbdays.nc"
            ncdfout=NetCDFFile(fileout, "w")
            
            write_lonlat(ncdfout,ncdfin)
            
            ncdfout.createDimension('time', None)
            if "reftime" in ncdfin.variables.keys():
                timevar="reftime"
            elif "time" in ncdfin.variables.keys():
                timevar="time"
            else:
                print "no time variable found"
                sys.exit(0)
            #create the reftime variable
            reftime=ncdfout.createVariable(timevar, "i", ('time',))
            reftime[:] = [0]
            reftimein = ncdfin.variables[timevar]
            reftime.units = reftimein.units
            reftime.standard_name = "forecast_reference_time"
            reftime.longname = "forecast reference time"
            reftime.calendar=reftimein.calendar

            ncdfout.createDimension('ensemble', nmember)
            
            var=ncdfout.createVariable(varname, 'f', ("time", "ensemble", "latitude", "longitude",))
            var[:]=varout
            ncdfout.close()
            ncdfin.close()
            filelistout.append(fileout)
        exp_stat=self.duplicate()
        exp_stat.set_filelist(filelistout)
        exp_stat.set_sdates(sdates)
        exp_stat.set_freq = relativedelta(months=+1)

        #get last date of the last file
        dateend=get_last_date(filelist[-1])
        period=str(sdates[0]).replace("-","")+"-"+str(dateend).replace("-","")

        #to be changed when final convention defined!!!!!
        exp_stat.set_varname(varname + "_nbdays_q" + str(percentile) + "_" + period)
        return exp_stat

     ######## seasonal_percentile ########


    def seasonal_percentile(self, percentile):
        """calculate the monthly percentile over the whole period
        for all file in filelist
        arguments:
        -percentile: percentile to be calculated 
        outputs:
        - object of type data with the computed index
        """

        cdo=Cdo()
        
        #get the needed variables from the object
        filelist=list(self.get_filelist())
        sdates=list(self.get_sdates())
        dataname=self.get_dataname()
        varname=self.get_varname()
        freq=self.get_freq()
        
        for i in range(len(filelist)):
            f=filelist[i]
            fileout=(f.split("/")[-1])
            fileout=fileout.split(".")[0]+"_q"+str(percentile)+".nc"
            cdo.timpctl(percentile, input=f+" -timmin "+ f+" -timmax "+f, 
                        output=fileout,
                        options="-r -f nc")

            filelist[i] = fileout

        
        exp_stat=self.duplicate()
        exp_stat.set_filelist(filelist)
        exp_stat.set_sdates(sdates)
        exp_stat.set_freq = relativedelta(months=+1)
        #to be changed when final convention defined!!!!!
        exp_stat.set_varname(varname + "_q" + str(percentile) )
       
        return exp_stat



     ######## daily statistics ########
    def cdo_statistic(self, stat, varout):
        """calcultate a cdo statistic: stat
        on all file of filelist
        -stat: statistics to be applied
        this statistic must have one only input
        which is the input file
        -name of the output varible
        """
        cdo=Cdo()
        self.see()
        #get the needed variables from the object
        filelist=self.get_filelist()
        sdates=list(self.get_sdates())
        dataname=self.get_dataname()
        varname=self.get_varname()
        freq=self.get_freq()
       
        for i in range(len(filelist)):
            f=filelist[i]
            fileout=(f.split("/")[-1])
            fileout=fileout.split(".")[0]+varout+".nc"
            cdo.chvar(varname,varout, 
                      input=" -"+stat+" "+f, output=fileout, Force=False)

            filelist[i] = fileout
        print filelist

        #create a copy of the  
        #self.set_filelist(filelist)
        self.set_freq(get_freq(filelist[0]))
        self.set_varname(varout)
        self.see()


    def calc_reg(self, indexname):
        """calculate an index for all files
        of the data object
        arguments:
        - name of the index 
        (has to be known by the mask_reg function)
        outputs:
        - object of type data with the computed index
        """

        self.see()
        #get the needed variables from the object
        filelist=self.get_filelist()
        sdates=list(self.get_sdates())
        dataname=self.get_dataname()
        varname=self.get_varname()
        freq=self.get_freq()
        
        exp_stat=self.duplicate()
        fileoutlst=[]
        for f in filelist:
            fout=compute_reg(f, dataname, indexname)
            fileoutlst.append(fout)
        
        exp_stat.set_sdates(sdates)
        exp_stat.set_filelist(fileoutlst)
        return(exp_stat)

    def calc_quantile(self, value, option):
        """estimate for file object which contain
        time serie (output of calc_reg) the quantile
        correponding to a given value. 
        if option quantile:
        arguments:
        - value: the value
        output:
        - the quantile associated to the value.
        if option value:
        arguments:
        - the quantile.
        output:
        - the value associated to a quantile.
        """
    
        cdo=Cdo()
        #get the needed variables from the object
        filelist=list(self.get_filelist())
        sdates=list(self.get_sdates())
        dataname=self.get_dataname()
        varname=self.get_varname()
        freq=self.get_freq()
        nmember=self.get_nmember()
        
    #check if there is no time overlap between the file
        for i in range(len(sdates)-1):
            lastdate = get_last_date(filelist[i])
            if (lastdate > sdates[i+1]):
                sdates[i]=sdates
                print "the file are overlaping"
                #sys.exit(0)
        merge=cdo.mergetime(input=filelist, options="-O -r -f nc")
        
        #open merge file
        ncdfmer=NetCDFFile(merge, "r")
        variables=ncdfmer.variables.keys()
        allDimNames = ncdfmer.dimensions.keys() 

        varmer = ncdfmer.variables[varname]
         
        shapmer=shape(varmer)
        varmer=numpy.resize(varmer,(shapmer[0]*shapmer[1]))
        shapmer=shape(varmer)

        if option == "quantile":
            aux=numpy.where(varmer>value)[0]
            print(aux)
            print(numpy.shape(aux))
            quantile=float((numpy.shape(aux))[0])/((numpy.shape(varmer))[0])
            print(numpy.shape(varmer))
            return(quantile)
        if option == "value":
            aux=numpy.percentile(varmer,value)
            return(aux)


   
    def calc_onset_thres(self, thres):
        """return the number of the first day of 
        the file of filelist exeeding the 
        """
    
        cdo=Cdo()
        #get the needed variables from the object
        filelist=list(self.get_filelist())
        sdates=list(self.get_sdates())
        dataname=self.get_dataname()
        varname=self.get_varname()
        freq=self.get_freq()
        nmember=self.get_nmember()     
        outputdir=self.get_outputdir()
        for ifile in range(len(filelist)):
            f=filelist[ifile]
            d=sdates[ifile]

            #calculate the onset
            onset=calc_fst_thres(f,dataname,thres)
            #convert to float to avoid trouble when writing
            onset=numpy.array(onset,  dtype='float32')

            varout="onset_"+varname+"_thres"
            dirout=outputdir+"/"+varout+"/"
            if (not os.path.isdir(dirout)):
                os.mkdir(dirout)
 
            fileout=dirout+varout+"_"+str(d).replace("-","")+".nc"
            ncdfout=NetCDFFile(fileout, "w")
            #ncdfin=NetCDFFile(f, "r")
            ncdfout.createDimension('ensemble', nmember)
            var=ncdfout.createVariable(varout, 'f', ("ensemble",))
            
            var.longname="onset date calculated over 1993 and 2009 for the two first month of the hindcast based on value in the model correponding to the quantile associated to 5mm/day in ERAint over this period"
            var.standard_name="onset"
            var.units="nb of day from 1st of May"
            if(shape(onset)==(1,)):
                var[:]=onset[0]
            else:
                var[:]=onset
            ncdfout.close()
             #generate a bug
            

